# Qhue is (c) Quentin Stafford-Fraser 2021
# but distributed under the GPL v2.
# It expects Python v3.

import re
import json

# for hostname retrieval for registering with the bridge
from socket import getfqdn

import requests

__all__ = ("Bridge", "QhueException", "create_new_username")

# default timeout in seconds
_DEFAULT_TIMEOUT = 5


class Resource(object):
    """
    A Resource represents an object or collection of objects in the Hue world,
    such as a light or a group of lights.
    It encapsulates an API method that can be called to examine or modify
    those objects, and makes it easy to construct the URLs needed.
    When you create a Resource, you are building a URL.
    When you call a Resource, you are making a request to that URL with some
    parameters.
    """
    def __init__(self, url, session, timeout=_DEFAULT_TIMEOUT, object_pairs_hook=None):
        self.url = url
        self.session = session
        self.address = url[url.find("/api"):]
        # Also find the bit after the username, if there is one
        self.short_address = None
        post_username_match = re.search(r"/api/[^/]*(.*)", url)
        if post_username_match is not None:
            self.short_address = post_username_match.group(1)
        self.timeout = timeout
        self.object_pairs_hook = object_pairs_hook

    def __call__(self, *args, **kwargs):
        # Preprocess args and kwargs
        url = self.url
        for a in args:
            url += "/" + str(a)
        http_method = kwargs.pop("http_method", "get" if not kwargs else "put").lower()

        # From each keyword, strip one trailing underscore if it exists,
        # then send them as parameters to the bridge. This allows for
        # "escaping" of keywords that might conflict with Python syntax
        # or with the specially-handled keyword "http_method".
        kwargs = {(k[:-1] if k.endswith("_") else k): v for k, v in kwargs.items()}
        if http_method == "put":
            r = self.session.put(url, data=json.dumps(kwargs, default=list), timeout=self.timeout)
        elif http_method == "post":
            r = self.session.post(url, data=json.dumps(kwargs, default=list), timeout=self.timeout)
        elif http_method == "delete":
            r = self.session.delete(url, timeout=self.timeout)
        else:
            r = self.session.get(url, timeout=self.timeout)
        if r.status_code != 200:
            raise QhueException("Received response {c} from {u}".format(c=r.status_code, u=url))
        resp = r.json(object_pairs_hook=self.object_pairs_hook)
        if type(resp) == list:
            # In theory, you can get more than one error from a single call
            # so they are returned as a list.
            errors = [m["error"] for m in resp if "error" in m]
            if errors:
                # In general, though, there will only be one error per call
                # so we return the type and address of the first one in the 
                # exception, to keep the exception type simple.
                raise QhueException(
                    message=",".join(e["description"] for e in errors),
                    type_id=",".join(str(e["type"]) for e in errors),
                    address=errors[0]['address']
                )
        return resp

    def __getattr__(self, name):
        return Resource(
            self.url + "/" + str(name),
            self.session,
            timeout=self.timeout,
            object_pairs_hook=self.object_pairs_hook,
        )

    __getitem__ = __getattr__


def _local_api_url(ip, username=None):
    if username is None:
        return "http://{}/api".format(ip)
    return "http://{}/api/{}".format(ip, username)


def create_new_username(ip, devicetype=None, timeout=_DEFAULT_TIMEOUT):
    """Interactive helper function to generate a new anonymous username.

    Args:
        ip: ip address of the bridge
        devicetype (optional): devicetype to register with the bridge. If
            unprovided, generates a device type based on the local hostname.
        timeout (optional, default=5): request timeout in seconds
    Raises:
        QhueException if something went wrong with username generation (for
            example, if the bridge button wasn't pressed).
    """
    res = Resource(_local_api_url(ip), requests.Session(), timeout)
    prompt = "Press the Bridge button, then press Return: "
    input(prompt)

    if devicetype is None:
        devicetype = "qhue#{}".format(getfqdn())

    # raises QhueException if something went wrong
    response = res(devicetype=devicetype, http_method="post")

    return response[0]["success"]["username"]


class Bridge(Resource):
    """
    A Bridge is a Resource that represents the top-level connection to a
    Philips Hue Bridge (or 'Hub').
    It is the basis for building other Resources that represent the things
    managed by that Bridge.
    """
    def __init__(self, ip, username, timeout=_DEFAULT_TIMEOUT, object_pairs_hook=None):
        """
        Create a new connection to a hue bridge.

        If a whitelisted username has not been generated yet, use
        create_new_username to have the bridge interactively generate
        a random username and then pass it to this function.

        Args:
            ip: ip address of the bridge
            username: valid username for the bridge
            timeout (optional, default=5): request timeout in seconds
            object_pairs_hook (optional): function called by JSON decoder with
                the result of any object literal as an ordered list of pairs.
        """
        self.ip = ip
        self.username = username
        url = _local_api_url(ip, username)
        self.session = requests.Session()
        super().__init__(url, self.session, timeout=timeout, object_pairs_hook=object_pairs_hook)


class QhueException(Exception):
    def __init__(self, message, type_id=None, address=None):
        self.message = message
        self.type_id = type_id
        self.address = address

        super().__init__(self.message)

    def __str__(self):
        return f'QhueException: {self.type_id} -> {self.message}'



# This is a basic solution for getting the access token from an OAuth 2 server.
#
# When you want to grant a piece of software access to an API, you often need to
# open a web browser, go to that API, authorise the access, and you are then
# redirected to a new URL with the appropriate access token passed in the
# request.
#
# This works fine for web apps, but command-line ones won't normally be
# exposing a URL to which OAuth can redirect. So we run a simple server that
# can be specified as the redirection address, and then make the token available
# when it comes through.
#
# OAuth 2 does require HTTPS, so we have to use a certificate.
# You can generate one and a key in the current directory with:
#
#    openssl req -x509 -nodes -newkey rsa:2048 -keyout key.pem -out cert.pem -days 365
#
# We will look for 'key.pem' and 'cert.pem' when serving.  This is a self-signed key
# so you'll probably need to tell your browser that it really is OK to go to this URL.
#
# You should specify 'https://localhost:8584' as the redirection address to the API.


from http.server import HTTPServer, BaseHTTPRequestHandler
import ssl

LOCAL_SERVER_PORT = 8584


class CollectorException(Exception):
    pass


class TokenReceivingServer(HTTPServer):
    """
    A simple HTTP server that listens on the specified port
    and stores the URL of the last request it received.
    """
    def __init__(self, port):
        self.received_request = None
        self.port = port
        print("Starting a small HTTP server to receive the callback")
        super().__init__(('', port), TokenHandler)

    def save_request(self, request: str):
        self.received_request = request

    def last_request(self) -> str:
        return self.received_request


class TokenHandler(BaseHTTPRequestHandler):
    """
    A little HTTP request handler which expects a token
    and stores it in the server.
    """
    def do_GET(self):
        self.server.save_request(self.path)
        self.send_response(200)
        self.send_header('Content-type', 'text/plain')
        self.end_headers()
        self.wfile.write(
            "Thank you.  Authentication token received. You can close this window.".encode()
        )


class TokenCollector():

    def __init__(self):
        self.http_server = TokenReceivingServer(LOCAL_SERVER_PORT)
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        context.load_cert_chain('cert.pem', 'key.pem')
        self.http_server.socket = context.wrap_socket(
            self.http_server.socket,
            server_side=True
        )

    def get_single_request(self):
        # Could make this wait until we have something that looks like a token
        print(f"Waiting for the callback on port {LOCAL_SERVER_PORT}...")
        while True:
            self.http_server.handle_request()
            req = self.http_server.last_request()
            if req is not None:
                break
        return f"https://localhost:{LOCAL_SERVER_PORT}{req}"




# Access to the Hue Hub when using the remote API
# via the Philips servers.
#
# Qhue is (c) Quentin Stafford-Fraser 2021
# but distributed under the GPL v2.
# It expects Python v3.

import webbrowser
import requests
from requests_oauthlib import OAuth2Session
from typing import Optional

from .qhue import Resource, _DEFAULT_TIMEOUT
from .oauth_receiver import TokenCollector

# Remote API root URL for use if outside the LAN
REMOTE_API_BASE = "https://api.meethue.com/bridge"

OAUTH_AUTHORIZE_URL = "https://api.meethue.com/v2/oauth2/authorize"
OAUTH_TOKEN_URL = "https://api.meethue.com/v2/oauth2/token"
OAUTH_REFRESH_URL = OAUTH_TOKEN_URL


def _remote_api_url(username):
    # The username is sometimes called a 'whitelist entry'
    # in the API docs. You can get one on your local LAN
    # using the create_new_username function as described in the
    # README.
    # TODO: We aren't yet dealing with the situation where you
    # don't have the username but are remote.
    return "{}/{}".format(REMOTE_API_BASE, username)


class RemoteBridge(Resource):
    """
    A RemoteBridge is a Resource that represents the top-level connection to a
    Philips Hue Bridge (or 'Hub') from outside that bridge's local network.
    It is the basis for building other Resources that represent the things
    managed by that Bridge.
    It is similar to a bridge, but uses OAuth authentication to the Philips server.
    """
    def __init__(self, username: str, timeout: float = _DEFAULT_TIMEOUT, object_pairs_hook=None):
        """
        Create a new connection to a remote Hue bridge.
        The 'username' is the same as for a local bridge -
        sometimes called a whitelist_identifier in the docs.
        It is not the user's identifier on the Philips site.
        """
        self.username = username
        self.session = requests.Session()
        url = _remote_api_url(username)
        super().__init__(url, self.session, timeout=timeout, object_pairs_hook=object_pairs_hook)

    def authorize(
        self,
        client_id: str,
        client_secret: str,
        token: Optional[str] = None,
        open_browser: bool = True,
        use_local_server: bool = False
    ):
        """
        Open a browser to the Hue site to ask you to authorise remote access.
        An existing token can be passed if available, otheriwse authorization will be needed:
        If open_browser is True, python will try to open your browser to the necessary URL,
        otherwise it will just print it out.
        Once authorised, this redirects to an HTTPS address, passing the required token.

        TODO: We don't currently refresh tokens.
        """
        # Use an oauth session in place of a standard requests session.
        self.session = OAuth2Session(client_id, token=token)
        if token is not None:
            return token
        authorization_url, state = self.session.authorization_url(OAUTH_AUTHORIZE_URL)
        if open_browser:
            print("Opening a browser to take you to", authorization_url)
            webbrowser.open_new(authorization_url)
        else:
            print("Open a browser at", authorization_url)

        if use_local_server:
            c = TokenCollector()
            redirect_response = c.get_single_request()

        else:
            redirect_response = input('Paste the full redirect URL here:')

        print("redirect response is ", redirect_response)

        return self.session.fetch_token(
            OAUTH_TOKEN_URL,
            client_secret=client_secret,
            authorization_response=redirect_response)
